/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright 2007 University of Washington
 * Copyright 2010 INRIA, Sophia Antipolis
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author:  Tom Henderson (tomhend@u.washington.edu)
 * Author:  Rao Naveed Bin Rais <Naveed-Bin.Rais@inria.fr>
 */
#include "ns3/address.h"
#include "ns3/address-utils.h"
#include "ns3/log.h"
#include "ns3/inet-socket-address.h"
#include "ns3/node.h"
#include "ns3/socket.h"
#include "ns3/udp-socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/udp-socket-factory.h"
#include "packet-sink.h"

/////////////Naveed/////////////
#include "ns3/seq-header.h"
#include <stdio.h>
/////////////Naveed/////////////

using namespace std;

namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("PacketSink");
NS_OBJECT_ENSURE_REGISTERED (PacketSink);

TypeId 
PacketSink::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::PacketSink")
    .SetParent<Application> ()
    .AddConstructor<PacketSink> ()
    .AddAttribute ("Local", "The Address on which to Bind the rx socket.",
                   AddressValue (),
                   MakeAddressAccessor (&PacketSink::m_local),
                   MakeAddressChecker ())
    .AddAttribute ("Protocol", "The type id of the protocol to use for the rx socket.",
                   TypeIdValue (UdpSocketFactory::GetTypeId ()),
                   MakeTypeIdAccessor (&PacketSink::m_tid),
                   MakeTypeIdChecker ())
    .AddTraceSource ("Rx", "A packet has been received",
                     MakeTraceSourceAccessor (&PacketSink::m_rxTrace))
    ;
  return tid;
}

PacketSink::PacketSink ()
{
  NS_LOG_FUNCTION (this);
  m_socket = 0;
  m_totalRx = 0;
  
  ////////////////////Naveed//////////////////////
  m_receiveCounter = 0;
  m_highPriority = 0;
  m_lowPriority = 0;
  m_prevSeqNumReceived = 0;
//  memset (m_packetIndex, 0, sizeof (m_packetIndex));
  memset (m_metrics, 0, sizeof (m_metrics));
  m_outputTimer = Timer::CANCEL_ON_DESTROY;
  m_outputTimer.SetFunction (&PacketSink::OutputTimerExpire, this);
  m_outputTimer.Schedule (Seconds (600.0));
  ////////////////////Naveed//////////////////////
}

PacketSink::~PacketSink()
{
  NS_LOG_FUNCTION (this);
}

uint32_t PacketSink::GetTotalRx() const
{
  return m_totalRx;
}
  
void PacketSink::DoDispose (void)
{
  NS_LOG_FUNCTION (this);
  m_socket = 0;
  m_socketList.clear ();

  // chain up
  Application::DoDispose ();
}


// Application Methods
void PacketSink::StartApplication()    // Called at time specified by Start
{
  NS_LOG_FUNCTION (this);
  // Create the socket if not already
  if (!m_socket)
    {
      m_socket = Socket::CreateSocket (GetNode(), m_tid);
      m_socket->Bind (m_local);
      m_socket->Listen ();
      m_socket->ShutdownSend ();
      if (addressUtils::IsMulticast (m_local))
        {
          Ptr<UdpSocket> udpSocket = DynamicCast<UdpSocket> (m_socket);
          if (udpSocket)
            {
              // equivalent to setsockopt (MCAST_JOIN_GROUP)
              udpSocket->MulticastJoinGroup (0, m_local);
            }
          else
            {
              NS_FATAL_ERROR ("Error: joining multicast on a non-UDP socket");
            }
        }
    }

  m_socket->SetRecvCallback (MakeCallback(&PacketSink::HandleRead, this));
  m_socket->SetAcceptCallback (
            MakeNullCallback<bool, Ptr<Socket>, const Address &> (),
            MakeCallback(&PacketSink::HandleAccept, this));
  m_socket->SetCloseCallbacks (
            MakeCallback(&PacketSink::HandlePeerClose, this),
            MakeCallback(&PacketSink::HandlePeerError, this));
}

void PacketSink::StopApplication()     // Called at time specified by Stop
{
  NS_LOG_FUNCTION (this);
  while(!m_socketList.empty()) //these are accepted sockets, close them
  {
    Ptr<Socket> acceptedSocket = m_socketList.front();
    m_socketList.pop_front();
    acceptedSocket->Close();
  }
  if (m_socket) 
    {
      m_socket->Close ();
      m_socket->SetRecvCallback (MakeNullCallback<void, Ptr<Socket> > ());
    }
    
  ////////////////////Naveed//////////////////////  
  double avgDelay = m_delay/m_receiveCounter;
//  Ptr<Node> node = this->GetObject<Node> ();
  std::cout << "Total received packets = " << m_receiveCounter << "\tTotal delay = " << avgDelay << " Seconds\n";
  ////////////////////Naveed//////////////////////
}

void PacketSink::HandleRead (Ptr<Socket> socket)
{
  NS_LOG_FUNCTION (this << socket);
  Ptr<Packet> packet;
  Address from;
  while (packet = socket->RecvFrom (from))
    {
      if (packet->GetSize() == 0)
        { //EOF
	  break;
        }
      if (InetSocketAddress::IsMatchingType (from))
        {
          InetSocketAddress address = InetSocketAddress::ConvertFrom (from);
          m_totalRx += packet->GetSize();
          NS_LOG_INFO ("Received " << packet->GetSize() << " bytes from " << 
                       address.GetIpv4() << " [" << address << "]" 
                       << " total Rx " << m_totalRx);
        }
        
      //////////////////////Naveed///////////////////////
      SeqNumHeader seqHdr;
      packet->RemoveHeader (seqHdr);
      int seqNum = seqHdr.GetSeqNum ();
      bool priority = seqHdr.GetPriority ();
      if (m_metrics[seqNum].m_packetIndex != 1)
        {
          m_metrics[seqNum].m_packetIndex = 1;
          m_prevSeqNumReceived = seqNum;
          Time time = seqHdr.GetSendTime ();
          if (priority)
            {
              m_highPriority++;
            }
          else
            {
              m_lowPriority++;
            }
          m_receiveCounter++;	      
          std::cout << "&&&&&&&&&&&&&& PacketSink::HandleRead, " << m_local << " Packet arrived with time = " << time.GetSeconds () << ", seq number = " << seqNum << " --- total number of received packets is = " << m_receiveCounter << " &&&&&&&&&&&&&&\n";
          Time currTime = Simulator::Now ();
          double delay = currTime.GetSeconds () - time.GetSeconds ();
          m_metrics[seqNum].m_packetDelay = delay;
          m_delay += delay;
          std::cout << "Total Packet Delay is = " << delay << "\n";
        }
      else
        {
          std::cout << "PacketSink::HandleRead, A duplicate copy of packet with sequence number " << seqNum << " arrived\n";
        }
	    //////////////////////Naveed///////////////////////
      
      m_rxTrace (packet, from);
    }
}

void PacketSink::HandlePeerClose (Ptr<Socket> socket)
{
  NS_LOG_INFO("PktSink, peerClose");
}
 
void PacketSink::HandlePeerError (Ptr<Socket> socket)
{
  NS_LOG_INFO("PktSink, peerError");
}
 

void PacketSink::HandleAccept (Ptr<Socket> s, const Address& from)
{
  NS_LOG_FUNCTION (this << s << from);
  s->SetRecvCallback (MakeCallback(&PacketSink::HandleRead, this));
  m_socketList.push_back (s);
}

//////////////////////Naveed///////////////////////
void
PacketSink::OutputTimerExpire (void)
{
  double avgDelay = m_delay/m_receiveCounter;
  std::cout << m_local << ", Packets = " << m_receiveCounter << "\t Delay = " << avgDelay << ", HP = " << m_highPriority << ", LP = " << m_lowPriority << "\n";
  m_outputTimer.Schedule (Seconds (20.0));
  
  FILE *fptr;
  char filename[100];
  int nodeid = GetNode ()->GetId ();
  sprintf (filename, "sta-%d.res", nodeid);
  fptr = fopen (filename, "w");

  for (int i=0; i<1000; i++)
    {
      fprintf (fptr, "%d\t%f\n", m_metrics[i].m_packetIndex, m_metrics[i].m_packetDelay);
    }
  fclose (fptr);
}
//////////////////////Naveed///////////////////////

} // Namespace ns3
