/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2006 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 * Author: Rao Naveed Bin Rais <Naveed-Bin.Rais@inria.fr>
 */
#include "ns3/packet.h"
#include "ns3/log.h"
#include "ns3/node.h"
#include "ns3/net-device.h"
#include "ns3/object-vector.h"
#include "ns3/trace-source-accessor.h"

///////////////Naveed/////////////////
#include "ns3/mac48-address.h"
///////////////Naveed/////////////////

#include "ipv4-l3-protocol.h"
#include "arp-l3-protocol.h"
#include "arp-header.h"
#include "arp-cache.h"
#include "ipv4-interface.h"

NS_LOG_COMPONENT_DEFINE ("ArpL3Protocol");

namespace ns3 {

const uint16_t ArpL3Protocol::PROT_NUMBER = 0x0806;

NS_OBJECT_ENSURE_REGISTERED (ArpL3Protocol);

TypeId 
ArpL3Protocol::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::ArpL3Protocol")
    .SetParent<Object> ()
    .AddConstructor<ArpL3Protocol> ()
    .AddAttribute ("CacheList",
                   "The list of ARP caches",
                   ObjectVectorValue (),
                   MakeObjectVectorAccessor (&ArpL3Protocol::m_cacheList),
                   MakeObjectVectorChecker<ArpCache> ())
    .AddTraceSource ("Drop",
                     "Packet dropped because not enough room in pending queue for a specific cache entry.",
                     MakeTraceSourceAccessor (&ArpL3Protocol::m_dropTrace))
    ;
  return tid;
}

ArpL3Protocol::ArpL3Protocol ()
{
  NS_LOG_FUNCTION (this);
}

ArpL3Protocol::~ArpL3Protocol ()
{
  NS_LOG_FUNCTION (this);
}

void 
ArpL3Protocol::SetNode (Ptr<Node> node)
{
  NS_LOG_FUNCTION (this);
  m_node = node;
}

/*
 * This method is called by AddAgregate and completes the aggregation
 * by setting the node in the ipv4 stack
 */
void
ArpL3Protocol::NotifyNewAggregate ()
{
  if (m_node == 0)
    {
      Ptr<Node>node = this->GetObject<Node> ();
      //verify that it's a valid node and that
      //the node was not set before
      if (node != 0)
        {
          this->SetNode (node);
        }
    }
  Object::NotifyNewAggregate ();
}

void 
ArpL3Protocol::DoDispose (void)
{
  NS_LOG_FUNCTION (this);
  for (CacheList::iterator i = m_cacheList.begin (); i != m_cacheList.end (); ++i)
    {
      Ptr<ArpCache> cache = *i;
      cache->Dispose ();
    }
  m_cacheList.clear ();
  m_node = 0;
  Object::DoDispose ();
}

Ptr<ArpCache> 
ArpL3Protocol::CreateCache (Ptr<NetDevice> device, Ptr<Ipv4Interface> interface)
{
  NS_LOG_FUNCTION (this << device << interface);
  Ptr<Ipv4L3Protocol> ipv4 = m_node->GetObject<Ipv4L3Protocol> ();
  Ptr<ArpCache> cache = CreateObject<ArpCache> ();
  cache->SetDevice (device, interface);
  NS_ASSERT (device->IsBroadcast ());
//  device->AddLinkChangeCallback (MakeCallback (&ArpCache::Flush, cache));

  /////////////////// Naveed ///////////////////
  std::cout << "Initializing Cache\n";
  InitializeARPCache(cache);
  /////////////////// Naveed ///////////////////
  
  cache->SetArpRequestCallback (MakeCallback (&ArpL3Protocol::SendArpRequest, this));
  m_cacheList.push_back (cache);
  return cache;
}

/////////////////// Naveed /////////////////////////////////////
void 
ArpL3Protocol::InitializeARPCache (Ptr<ArpCache> cache)
{
  AddEntryToArpTable("192.168.0.0", "00:00:00:00:00:00", cache);
  AddEntryToArpTable("10.1.3.1", "00:00:00:00:00:01", cache);
  AddEntryToArpTable("10.1.3.2", "00:00:00:00:00:02", cache);
  AddEntryToArpTable("12.1.3.1", "00:00:00:00:00:03", cache);
  AddEntryToArpTable("12.1.3.2", "00:00:00:00:00:04", cache);
  AddEntryToArpTable("14.1.3.1", "00:00:00:00:00:05", cache);
  AddEntryToArpTable("14.1.3.2", "00:00:00:00:00:06", cache);
  AddEntryToArpTable("192.168.0.1", "00:00:00:00:00:07", cache);
  AddEntryToArpTable("192.168.0.2", "00:00:00:00:00:08", cache);
  AddEntryToArpTable("192.168.0.3", "00:00:00:00:00:09", cache);
  AddEntryToArpTable("192.168.0.4", "00:00:00:00:00:0A", cache);
  AddEntryToArpTable("192.168.0.5", "00:00:00:00:00:0B", cache);
  AddEntryToArpTable("192.168.0.6", "00:00:00:00:00:0C", cache);
  AddEntryToArpTable("192.168.0.7", "00:00:00:00:00:0D", cache);
  AddEntryToArpTable("192.168.6.1", "00:00:00:00:00:0E", cache);
  AddEntryToArpTable("192.168.6.2", "00:00:00:00:00:0F", cache);
  AddEntryToArpTable("192.168.6.3", "00:00:00:00:00:10", cache);
/*
  AddEntryToArpTable("18.1.3.1", "00:00:00:00:00:11", cache);
  AddEntryToArpTable("18.1.3.2", "00:00:00:00:00:12", cache);
  AddEntryToArpTable("19.1.3.1", "00:00:00:00:00:13", cache);
  AddEntryToArpTable("19.1.3.2", "00:00:00:00:00:14", cache);
  AddEntryToArpTable("20.1.3.1", "00:00:00:00:00:15", cache);
  AddEntryToArpTable("20.1.3.2", "00:00:00:00:00:16", cache);
  AddEntryToArpTable("21.1.3.1", "00:00:00:00:00:17", cache);
  AddEntryToArpTable("21.1.3.2", "00:00:00:00:00:18", cache);
  AddEntryToArpTable("22.1.3.1", "00:00:00:00:00:19", cache);
  AddEntryToArpTable("22.1.3.2", "00:00:00:00:00:1A", cache);
  AddEntryToArpTable("23.1.3.1", "00:00:00:00:00:1B", cache);
  AddEntryToArpTable("23.1.3.2", "00:00:00:00:00:1C", cache);
  AddEntryToArpTable("24.1.3.1", "00:00:00:00:00:1D", cache);
  AddEntryToArpTable("24.1.3.2", "00:00:00:00:00:1E", cache);
  AddEntryToArpTable("25.1.3.1", "00:00:00:00:00:1F", cache);
  AddEntryToArpTable("25.1.3.2", "00:00:00:00:00:20", cache);
  AddEntryToArpTable("26.1.3.1", "00:00:00:00:00:21", cache);
  AddEntryToArpTable("26.1.3.2", "00:00:00:00:00:22", cache);
  AddEntryToArpTable("27.1.3.1", "00:00:00:00:00:23", cache);
  AddEntryToArpTable("27.1.3.2", "00:00:00:00:00:24", cache);
  AddEntryToArpTable("28.1.3.1", "00:00:00:00:00:25", cache);
  AddEntryToArpTable("28.1.3.2", "00:00:00:00:00:26", cache);
  AddEntryToArpTable("29.1.3.1", "00:00:00:00:00:27", cache);
  AddEntryToArpTable("29.1.3.2", "00:00:00:00:00:28", cache);
  AddEntryToArpTable("30.1.3.1", "00:00:00:00:00:29", cache);
  AddEntryToArpTable("30.1.3.2", "00:00:00:00:00:2A", cache);
  AddEntryToArpTable("31.1.3.1", "00:00:00:00:00:2B", cache);
  AddEntryToArpTable("31.1.3.2", "00:00:00:00:00:2C", cache);
  AddEntryToArpTable("32.1.3.1", "00:00:00:00:00:2D", cache);
  AddEntryToArpTable("32.1.3.2", "00:00:00:00:00:2E", cache);
  AddEntryToArpTable("33.1.3.1", "00:00:00:00:00:2F", cache);
  AddEntryToArpTable("33.1.3.2", "00:00:00:00:00:30", cache);
  AddEntryToArpTable("34.1.3.1", "00:00:00:00:00:31", cache);
  AddEntryToArpTable("34.1.3.2", "00:00:00:00:00:32", cache);
  AddEntryToArpTable("35.1.3.1", "00:00:00:00:00:33", cache);
  AddEntryToArpTable("35.1.3.2", "00:00:00:00:00:34", cache);
  AddEntryToArpTable("36.1.3.1", "00:00:00:00:00:35", cache);
  AddEntryToArpTable("36.1.3.2", "00:00:00:00:00:36", cache);
  AddEntryToArpTable("37.1.3.1", "00:00:00:00:00:37", cache);
  AddEntryToArpTable("37.1.3.2", "00:00:00:00:00:38", cache);
  AddEntryToArpTable("38.1.3.1", "00:00:00:00:00:39", cache);
  AddEntryToArpTable("38.1.3.2", "00:00:00:00:00:3A", cache);
  AddEntryToArpTable("39.1.3.1", "00:00:00:00:00:3B", cache);
  AddEntryToArpTable("39.1.3.2", "00:00:00:00:00:3C", cache);
  AddEntryToArpTable("40.1.3.1", "00:00:00:00:00:3D", cache);
  AddEntryToArpTable("40.1.3.2", "00:00:00:00:00:3E", cache);
  AddEntryToArpTable("41.1.3.1", "00:00:00:00:00:3F", cache);
  AddEntryToArpTable("41.1.3.2", "00:00:00:00:00:40", cache);
  AddEntryToArpTable("42.1.3.1", "00:00:00:00:00:41", cache);
  AddEntryToArpTable("42.1.3.2", "00:00:00:00:00:42", cache);
  AddEntryToArpTable("43.1.3.1", "00:00:00:00:00:43", cache);
  AddEntryToArpTable("43.1.3.2", "00:00:00:00:00:44", cache);
  AddEntryToArpTable("44.1.3.1", "00:00:00:00:00:45", cache);
  AddEntryToArpTable("44.1.3.2", "00:00:00:00:00:46", cache);
  AddEntryToArpTable("45.1.3.1", "00:00:00:00:00:47", cache);
  AddEntryToArpTable("45.1.3.2", "00:00:00:00:00:48", cache);
  AddEntryToArpTable("10.1.1.1", "00:00:00:00:00:49", cache);
  AddEntryToArpTable("10.1.1.2", "00:00:00:00:00:4A", cache);
  AddEntryToArpTable("10.1.1.3", "00:00:00:00:00:4B", cache);
  AddEntryToArpTable("10.1.1.4", "00:00:00:00:00:4C", cache);
  AddEntryToArpTable("10.1.1.5", "00:00:00:00:00:4D", cache);
  AddEntryToArpTable("10.1.1.6", "00:00:00:00:00:4E", cache);
  AddEntryToArpTable("10.1.1.7", "00:00:00:00:00:4F", cache);
  AddEntryToArpTable("10.1.1.8", "00:00:00:00:00:50", cache);
  AddEntryToArpTable("10.1.1.9", "00:00:00:00:00:51", cache);
  AddEntryToArpTable("10.1.1.10", "00:00:00:00:00:52", cache);
  AddEntryToArpTable("10.1.1.11", "00:00:00:00:00:53", cache);
  AddEntryToArpTable("10.1.1.12", "00:00:00:00:00:54", cache);
  AddEntryToArpTable("10.1.1.13", "00:00:00:00:00:55", cache);
  AddEntryToArpTable("10.1.1.14", "00:00:00:00:00:56", cache);
  AddEntryToArpTable("10.1.1.15", "00:00:00:00:00:57", cache);
  AddEntryToArpTable("10.1.1.16", "00:00:00:00:00:58", cache);
  AddEntryToArpTable("10.1.1.17", "00:00:00:00:00:59", cache);
  AddEntryToArpTable("10.1.1.18", "00:00:00:00:00:5A", cache);
  AddEntryToArpTable("10.1.1.19", "00:00:00:00:00:5B", cache);
  AddEntryToArpTable("10.1.1.20", "00:00:00:00:00:5C", cache);
  AddEntryToArpTable("10.1.1.21", "00:00:00:00:00:5D", cache);
  AddEntryToArpTable("10.1.1.22", "00:00:00:00:00:5E", cache);
  AddEntryToArpTable("10.1.1.23", "00:00:00:00:00:5F", cache);
  AddEntryToArpTable("10.1.1.24", "00:00:00:00:00:60", cache);
  AddEntryToArpTable("10.1.1.25", "00:00:00:00:00:61", cache);
  AddEntryToArpTable("10.1.1.26", "00:00:00:00:00:62", cache);
  AddEntryToArpTable("10.1.1.27", "00:00:00:00:00:63", cache);
  AddEntryToArpTable("10.1.1.28", "00:00:00:00:00:64", cache);
  AddEntryToArpTable("10.1.1.29", "00:00:00:00:00:65", cache);
  AddEntryToArpTable("10.1.1.30", "00:00:00:00:00:66", cache);
  AddEntryToArpTable("10.1.1.31", "00:00:00:00:00:67", cache);
  AddEntryToArpTable("10.1.1.32", "00:00:00:00:00:68", cache);
  AddEntryToArpTable("10.1.1.33", "00:00:00:00:00:69", cache);
  AddEntryToArpTable("10.1.1.34", "00:00:00:00:00:6A", cache);
  AddEntryToArpTable("10.1.1.35", "00:00:00:00:00:6B", cache);
  AddEntryToArpTable("10.1.1.36", "00:00:00:00:00:6C", cache);
  AddEntryToArpTable("10.1.1.37", "00:00:00:00:00:6D", cache);
  AddEntryToArpTable("10.1.1.38", "00:00:00:00:00:6E", cache);
  AddEntryToArpTable("10.1.1.39", "00:00:00:00:00:6F", cache);
  AddEntryToArpTable("10.1.1.40", "00:00:00:00:00:70", cache);
  AddEntryToArpTable("10.1.1.41", "00:00:00:00:00:71", cache);
  AddEntryToArpTable("10.1.1.42", "00:00:00:00:00:72", cache);
  AddEntryToArpTable("10.1.1.43", "00:00:00:00:00:73", cache);
  AddEntryToArpTable("10.1.1.44", "00:00:00:00:00:74", cache);
  AddEntryToArpTable("10.1.1.45", "00:00:00:00:00:75", cache);
  AddEntryToArpTable("10.1.1.46", "00:00:00:00:00:76", cache);
  AddEntryToArpTable("10.1.1.47", "00:00:00:00:00:77", cache);
  AddEntryToArpTable("10.1.1.48", "00:00:00:00:00:78", cache);
  AddEntryToArpTable("10.1.1.49", "00:00:00:00:00:79", cache);
  AddEntryToArpTable("192.168.6.1", "00:00:00:00:00:7A", cache);
  AddEntryToArpTable("192.168.6.2", "00:00:00:00:00:7B", cache);
  AddEntryToArpTable("192.168.6.3", "00:00:00:00:00:7C", cache);
  AddEntryToArpTable("192.168.6.4", "00:00:00:00:00:7D", cache);
  AddEntryToArpTable("192.168.6.5", "00:00:00:00:00:7E", cache);
  AddEntryToArpTable("192.168.6.6", "00:00:00:00:00:7F", cache);
  AddEntryToArpTable("192.168.6.7", "00:00:00:00:00:80", cache);
  AddEntryToArpTable("192.168.6.8", "00:00:00:00:00:81", cache);
  AddEntryToArpTable("192.168.6.9", "00:00:00:00:00:82", cache);
  AddEntryToArpTable("192.168.6.10", "00:00:00:00:00:83", cache);
  AddEntryToArpTable("192.168.6.11", "00:00:00:00:00:84", cache);
  AddEntryToArpTable("192.168.6.12", "00:00:00:00:00:85", cache);
  AddEntryToArpTable("192.168.6.13", "00:00:00:00:00:86", cache);
  AddEntryToArpTable("192.168.6.14", "00:00:00:00:00:87", cache);
  AddEntryToArpTable("192.168.6.15", "00:00:00:00:00:88", cache);
  AddEntryToArpTable("192.168.6.16", "00:00:00:00:00:89", cache);
  AddEntryToArpTable("192.168.6.17", "00:00:00:00:00:8A", cache);
  AddEntryToArpTable("192.168.6.18", "00:00:00:00:00:8B", cache);
  AddEntryToArpTable("192.168.6.19", "00:00:00:00:00:8C", cache);
  AddEntryToArpTable("192.168.6.20", "00:00:00:00:00:8D", cache);
  AddEntryToArpTable("192.168.6.21", "00:00:00:00:00:8E", cache);
  AddEntryToArpTable("192.168.6.22", "00:00:00:00:00:8F", cache);
  AddEntryToArpTable("192.168.6.23", "00:00:00:00:00:90", cache);
  AddEntryToArpTable("192.168.6.24", "00:00:00:00:00:91", cache);
  AddEntryToArpTable("192.168.6.25", "00:00:00:00:00:92", cache);
  AddEntryToArpTable("192.168.6.26", "00:00:00:00:00:93", cache);
  AddEntryToArpTable("192.168.6.27", "00:00:00:00:00:94", cache);
  AddEntryToArpTable("192.168.6.28", "00:00:00:00:00:95", cache);
  AddEntryToArpTable("192.168.6.29", "00:00:00:00:00:96", cache);
  AddEntryToArpTable("192.168.6.30", "00:00:00:00:00:97", cache);
  AddEntryToArpTable("192.168.6.31", "00:00:00:00:00:98", cache);
  AddEntryToArpTable("192.168.6.32", "00:00:00:00:00:99", cache);
  AddEntryToArpTable("192.168.6.33", "00:00:00:00:00:9A", cache);
  AddEntryToArpTable("192.168.6.34", "00:00:00:00:00:9B", cache);
  AddEntryToArpTable("192.168.6.35", "00:00:00:00:00:9C", cache);
  AddEntryToArpTable("192.168.6.36", "00:00:00:00:00:9D", cache);
  AddEntryToArpTable("192.168.6.37", "00:00:00:00:00:9E", cache);
  AddEntryToArpTable("192.168.6.38", "00:00:00:00:00:9F", cache);
  AddEntryToArpTable("192.168.6.39", "00:00:00:00:00:A0", cache);
  AddEntryToArpTable("192.168.6.40", "00:00:00:00:00:A1", cache);
  AddEntryToArpTable("192.168.8.1", "00:00:00:00:00:A2", cache);
  AddEntryToArpTable("192.168.8.2", "00:00:00:00:00:A3", cache);
  AddEntryToArpTable("192.168.8.3", "00:00:00:00:00:A4", cache);
  AddEntryToArpTable("192.168.8.4", "00:00:00:00:00:A5", cache);
  AddEntryToArpTable("192.168.8.5", "00:00:00:00:00:A6", cache);
  AddEntryToArpTable("192.168.8.6", "00:00:00:00:00:A7", cache);
  AddEntryToArpTable("192.168.8.7", "00:00:00:00:00:A8", cache);
  AddEntryToArpTable("192.168.8.8", "00:00:00:00:00:A9", cache);
  AddEntryToArpTable("192.168.8.9", "00:00:00:00:00:AA", cache);
  AddEntryToArpTable("192.168.8.10", "00:00:00:00:00:AB", cache);
  AddEntryToArpTable("192.168.8.11", "00:00:00:00:00:AC", cache);
  AddEntryToArpTable("192.168.8.12", "00:00:00:00:00:AD", cache);
  AddEntryToArpTable("192.168.8.13", "00:00:00:00:00:AE", cache);
  AddEntryToArpTable("192.168.8.14", "00:00:00:00:00:AF", cache);
  AddEntryToArpTable("192.168.8.15", "00:00:00:00:00:B0", cache);
  AddEntryToArpTable("192.168.8.16", "00:00:00:00:00:B1", cache);
  AddEntryToArpTable("192.168.8.17", "00:00:00:00:00:B2", cache);
  AddEntryToArpTable("192.168.8.18", "00:00:00:00:00:B3", cache);
  AddEntryToArpTable("192.168.8.19", "00:00:00:00:00:B4", cache);
  AddEntryToArpTable("192.168.8.20", "00:00:00:00:00:B5", cache);
  AddEntryToArpTable("192.168.8.21", "00:00:00:00:00:B6", cache);
  AddEntryToArpTable("192.168.8.22", "00:00:00:00:00:B7", cache);
  AddEntryToArpTable("192.168.8.23", "00:00:00:00:00:B8", cache);
  AddEntryToArpTable("192.168.8.24", "00:00:00:00:00:B9", cache);
  AddEntryToArpTable("192.168.8.25", "00:00:00:00:00:BA", cache);
  AddEntryToArpTable("192.168.8.26", "00:00:00:00:00:BB", cache);
  AddEntryToArpTable("192.168.8.27", "00:00:00:00:00:BC", cache);
  AddEntryToArpTable("192.168.8.28", "00:00:00:00:00:BD", cache);
  AddEntryToArpTable("192.168.8.29", "00:00:00:00:00:BE", cache);
  AddEntryToArpTable("192.168.8.30", "00:00:00:00:00:BF", cache);
  AddEntryToArpTable("192.168.8.31", "00:00:00:00:00:C0", cache);
  AddEntryToArpTable("192.168.8.32", "00:00:00:00:00:C1", cache);
  AddEntryToArpTable("192.168.8.33", "00:00:00:00:00:C2", cache);
  AddEntryToArpTable("192.168.8.34", "00:00:00:00:00:C3", cache);
  AddEntryToArpTable("192.168.8.35", "00:00:00:00:00:C4", cache);
  AddEntryToArpTable("192.168.8.36", "00:00:00:00:00:C5", cache);
  AddEntryToArpTable("192.168.8.37", "00:00:00:00:00:C6", cache);
  AddEntryToArpTable("192.168.8.38", "00:00:00:00:00:C7", cache);
  AddEntryToArpTable("192.168.8.39", "00:00:00:00:00:C8", cache);
  AddEntryToArpTable("192.168.8.40", "00:00:00:00:00:C9", cache);
*/
}                                                    
                                                     
void                                                 
ArpL3Protocol::AddEntryToArpTable(const char* destIp, const char* destMac, Ptr<ArpCache> cache)
{                                                    
  ArpCache::Entry *entry;                            
  Ipv4Address destIpAddress(destIp);                 
  Mac48Address macAddress(destMac);                  
                                                     
  entry = cache->Add (destIpAddress);                
  entry->SetDstMacAddress((Address)macAddress);      
}                                                    
                                                     
Ipv4Address                                          
ArpL3Protocol::SearchForIpv4Address (Address address)
{                                                    
  Ptr<ArpCache> cache;                               
  Ipv4Address ipv4Address;                           
                                                     
  for (CacheList::iterator i = m_cacheList.begin (); i != m_cacheList.end (); ++i)
    {                                                
      cache = *i;                                    
      ipv4Address = cache->SearchForIpv4Address (address);
      if (ipv4Address != "127.0.0.1")                
        {                                            
	  return ipv4Address;                              
	}
    }
  
  return "127.0.0.1";
}

Address
ArpL3Protocol::SearchForMacAddress (Ipv4Address ipv4Address)
{
  Ptr<ArpCache> cache;
  Address address;
  
  for (CacheList::iterator i = m_cacheList.begin (); i != m_cacheList.end (); ++i)
    {
      cache = *i;
      address = cache->SearchForMacAddress (ipv4Address);
      return address;
    }
  
  return address;
}
/////////////////// Naveed /////////////////////////////////////

Ptr<ArpCache>
ArpL3Protocol::FindCache (Ptr<NetDevice> device)
{
  NS_LOG_FUNCTION (this << device);
  for (CacheList::const_iterator i = m_cacheList.begin (); i != m_cacheList.end (); i++)
    {
      if ((*i)->GetDevice () == device)
	{
	  return *i;
	}
    }
  NS_ASSERT (false);
  // quiet compiler
  return 0;
}

void 
ArpL3Protocol::Receive(Ptr<NetDevice> device, Ptr<const Packet> p, uint16_t protocol, const Address &from,
                       const Address &to, NetDevice::PacketType packetType)
{
  NS_LOG_FUNCTION (this << device << p->GetSize () << protocol << from << to << packetType);

  Ptr<Packet> packet = p->Copy ();

  NS_LOG_LOGIC ("ARP: received packet of size "<< packet->GetSize ());

  Ptr<ArpCache> cache = FindCache (device);

  // 
  // If we're connected to a real world network, then some of the fields sizes 
  // in an ARP packet can vary in ways not seen in simulations.  We need to be
  // able to detect ARP packets with headers we don't recongnize and not process
  // them instead of crashing.  The ArpHeader will return 0 if it can't deal
  // with the received header.
  //
  ArpHeader arp;
  uint32_t size = packet->RemoveHeader (arp);
  if (size == 0)
    {
      NS_LOG_LOGIC ("ARP: Cannot remove ARP header");
      return;
    }
  NS_LOG_LOGIC ("ARP: received "<< (arp.IsRequest ()? "request" : "reply") <<
            " node="<<m_node->GetId ()<<", got request from " <<
            arp.GetSourceIpv4Address () << " for address " <<
            arp.GetDestinationIpv4Address () << "; we have addresses: ");
  for (uint32_t i = 0; i < cache->GetInterface ()->GetNAddresses (); i++)
    {
      NS_LOG_LOGIC (cache->GetInterface ()->GetAddress (i).GetLocal () << ", ");
    }

  /**
   * Note: we do not update the ARP cache when we receive an ARP request
   * from an unknown node. See bug #107
   */
  bool found = false;
  for (uint32_t i = 0; i < cache->GetInterface ()->GetNAddresses (); i++)
    {
      if (arp.IsRequest () && arp.GetDestinationIpv4Address () == 
        cache->GetInterface ()->GetAddress (i).GetLocal ()) 
        {
          found = true;
          NS_LOG_LOGIC ("node="<<m_node->GetId () <<", got request from " << 
                arp.GetSourceIpv4Address () << " -- send reply");
          SendArpReply (cache, arp.GetDestinationIpv4Address (), arp.GetSourceIpv4Address (),
                    arp.GetSourceHardwareAddress ());
          break;
        } 
      else if (arp.IsReply () && 
        arp.GetDestinationIpv4Address ().IsEqual (cache->GetInterface ()->GetAddress (i).GetLocal ()) &&
        arp.GetDestinationHardwareAddress () == device->GetAddress ()) 
        {
          found = true;
          Ipv4Address from = arp.GetSourceIpv4Address ();
          ArpCache::Entry *entry = cache->Lookup (from);
          if (entry != 0)
            {
              if (entry->IsWaitReply ()) 
                {
                  NS_LOG_LOGIC ("node="<< m_node->GetId () << 
                    ", got reply from " << arp.GetSourceIpv4Address ()
                     << " for waiting entry -- flush");
                  Address from_mac = arp.GetSourceHardwareAddress ();
                  entry->MarkAlive (from_mac);
                  Ptr<Packet> pending = entry->DequeuePending();
                  while (pending != 0)
                    {
                      cache->GetInterface ()->Send (pending,
                                                arp.GetSourceIpv4Address ());
                      pending = entry->DequeuePending();
                    }
                } 
              else 
                {
                  // ignore this reply which might well be an attempt 
                  // at poisening my arp cache.
                  NS_LOG_LOGIC("node="<<m_node->GetId ()<<", got reply from " << 
                        arp.GetSourceIpv4Address () << 
                        " for non-waiting entry -- drop");
                  m_dropTrace (packet);
                }
            } 
          else 
            {
              NS_LOG_LOGIC ("node="<<m_node->GetId ()<<", got reply for unknown entry -- drop");
              m_dropTrace (packet);
            }
          break;
        }
    }
  if (found == false)
    {
      NS_LOG_LOGIC ("node="<<m_node->GetId ()<<", got request from " <<
                arp.GetSourceIpv4Address () << " for unknown address " <<
                arp.GetDestinationIpv4Address () << " -- drop");
    }
}

bool 
ArpL3Protocol::Lookup (Ptr<Packet> packet, Ipv4Address destination, 
                       Ptr<NetDevice> device,
                       Ptr<ArpCache> cache,
                       Address *hardwareDestination)
{
  NS_LOG_FUNCTION (this << packet << destination << device << cache);
  ArpCache::Entry *entry = cache->Lookup (destination);
  if (entry != 0)
    {
      if (entry->IsExpired ()) 
        {
          if (entry->IsDead ()) 
            {
              NS_LOG_LOGIC ("node="<<m_node->GetId ()<<
                        ", dead entry for " << destination << " expired -- send arp request");
              entry->MarkWaitReply (packet);
              SendArpRequest (cache, destination);
            } 
          else if (entry->IsAlive ()) 
            {
              NS_LOG_LOGIC ("node="<<m_node->GetId ()<<
                        ", alive entry for " << destination << " expired -- send arp request");
              entry->MarkWaitReply (packet);
              SendArpRequest (cache, destination);
            } 
          else if (entry->IsWaitReply ()) 
            {
              NS_FATAL_ERROR ("Test for possibly unreachable code-- please file a bug report, with a test case, if this is ever hit");
            }
        } 
      else 
        {
          if (entry->IsDead ()) 
            {
              NS_LOG_LOGIC ("node="<<m_node->GetId ()<<
                            ", dead entry for " << destination << " valid -- drop");
              m_dropTrace (packet);
            } 
          else if (entry->IsAlive ()) 
            {
              NS_LOG_LOGIC ("node="<<m_node->GetId ()<<
                            ", alive entry for " << destination << " valid -- send");
	      *hardwareDestination = entry->GetMacAddress ();
              return true;
            } 
          else if (entry->IsWaitReply ()) 
            {
              NS_LOG_LOGIC ("node="<<m_node->GetId ()<<
                            ", wait reply for " << destination << " valid -- drop previous");
              if (!entry->UpdateWaitReply (packet))
                {
                  m_dropTrace (packet);
                }
            }
        }
    }
  else
    {
      // This is our first attempt to transmit data to this destination.
      NS_LOG_LOGIC ("node="<<m_node->GetId ()<<
                ", no entry for " << destination << " -- send arp request");
      entry = cache->Add (destination);
      entry->MarkWaitReply (packet);
      SendArpRequest (cache, destination);
    }
  return false;
}

void
ArpL3Protocol::SendArpRequest (Ptr<const ArpCache> cache, Ipv4Address to)
{
  NS_LOG_FUNCTION (this << cache << to);
  ArpHeader arp;
  // need to pick a source address; use routing implementation to select
  Ptr<Ipv4L3Protocol> ipv4 = m_node->GetObject<Ipv4L3Protocol> ();
  Ptr<NetDevice> device = cache->GetDevice ();
  NS_ASSERT (device != 0);
  Ipv4Header header;
  header.SetDestination (to);
  Ptr<Packet> packet = Create<Packet> ();
  Ipv4Address source = ipv4->SelectSourceAddress (device,  to, Ipv4InterfaceAddress::GLOBAL);
  NS_LOG_LOGIC ("ARP: sending request from node "<<m_node->GetId ()<<
            " || src: " << device->GetAddress () << " / " << source <<
            " || dst: " << device->GetBroadcast () << " / " << to);
  arp.SetRequest (device->GetAddress (), source, device->GetBroadcast (), to);
  packet->AddHeader (arp);
  cache->GetDevice ()->Send (packet, device->GetBroadcast (), PROT_NUMBER);
}

void
ArpL3Protocol::SendArpReply (Ptr<const ArpCache> cache, Ipv4Address myIp, Ipv4Address toIp, Address toMac)
{
  NS_LOG_FUNCTION (this << cache << toIp << toMac);
  ArpHeader arp;
  NS_LOG_LOGIC ("ARP: sending reply from node "<<m_node->GetId ()<<
            "|| src: " << cache->GetDevice ()->GetAddress () << 
            " / " << myIp <<
            " || dst: " << toMac << " / " << toIp);
  arp.SetReply (cache->GetDevice ()->GetAddress (), myIp, toMac, toIp);
  Ptr<Packet> packet = Create<Packet> ();
  packet->AddHeader (arp);
  cache->GetDevice ()->Send (packet, toMac, PROT_NUMBER);
}

}//namespace ns3
