/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Rao Naveed Bin Rais <Naveed-Bin.Rais@inria.fr>
 */
#include <cmath>
#include "ns3/simulator.h"
#include "ns3/random-variable.h"
#include "ns3/pointer.h"
#include "bonn-motion-mobility-model.h"
#include "position-allocator.h"
#include "ns3/double.h"

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (BonnMotionMobilityModel);

TypeId
BonnMotionMobilityModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::BonnMotionMobilityModel")
    .SetParent<MobilityModel> ()
    .SetGroupName ("Mobility")
    .AddConstructor<BonnMotionMobilityModel> ()
    .AddAttribute ("Speed",
                   "A random variable used to pick the speed of a random waypoint model.",
                   RandomVariableValue (UniformVariable (0.3, 0.7)),
                   MakeDoubleAccessor (&BonnMotionMobilityModel::m_speed),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("Pause",
                   "A random variable used to pick the pause of a random waypoint model.",
                   RandomVariableValue (ConstantVariable (2.0)),
                   MakeTimeAccessor (&BonnMotionMobilityModel::m_pause),
                   MakeTimeChecker ())
    .AddAttribute ("PositionAllocator",
                   "The position model used to pick a destination point.",
                   PointerValue (),
                   MakePointerAccessor (&BonnMotionMobilityModel::m_position),
                   MakePointerChecker<PositionAllocator> ());
  
  return tid;
}

BonnMotionMobilityModel::BonnMotionMobilityModel ()
{
  m_event = Simulator::ScheduleNow (&BonnMotionMobilityModel::BeginWalk, this);
}

void
BonnMotionMobilityModel::BeginWalk (void)
{
  m_helper.Update ();
  Vector m_current = m_helper.GetCurrentPosition ();
  Vector destination = m_position->GetNext ();
  Ptr<BonnMotionPositionAllocator> temp;
  temp = StaticCast<BonnMotionPositionAllocator> (m_position);
  Time time = temp->GetNextTime ();
  Time diff = time - m_prevTime;
  m_prevTime = time;
  
  double dx = (destination.x - m_current.x);
  double dy = (destination.y - m_current.y);
  double dz = (destination.z - m_current.z);

  if (dx == 0 && dy == 0)
    {
      m_helper.Update ();
      m_helper.Pause ();
      m_event = Simulator::Schedule (diff,
				 &BonnMotionMobilityModel::BeginWalk, this);
    }

  double output;
  if (diff.GetSeconds () == 0)
      output = 0;
  else
      output = CalculateDistance (destination, m_current) / diff.GetSeconds ();
  if (output < 0)
      m_speed = -output;
  else
      m_speed = output;
  double k = m_speed / std::sqrt (dx*dx + dy*dy + dz*dz);

  m_helper.SetVelocity (Vector (k*dx, k*dy, k*dz));
//  std::cout << "BeginWalk, m_current.x = " << m_current.x << ", m_current.y = " << m_current.y << "\n";
//  std::cout << "BeginWalk, destination.x = " << destination.x << ", destination.y = " << destination.y << "\n";
//  std::cout << "dx = " << dx << ", dy = " << dy << "\n";
//  std::cout << "Speed = " << m_speed << ", Velocity = " << Vector (k*dx, k*dy, k*dz) << "\n";
  m_helper.Unpause ();
//  Time travelDelay = Seconds (CalculateDistance (destination, m_current) / speed);
//  std::cout << "Current Time = " << time.GetSeconds () << ", Previous Time = " << m_prevTime.GetSeconds () << ", Next move is scheduled at " << diff.GetSeconds () << "\n";
  
  if (diff.IsPositive ())
    {
      m_event = Simulator::Schedule (diff,
				 &BonnMotionMobilityModel::BeginWalk, this);
    }
  NotifyCourseChange ();
}
/*
void
BonnMotionMobilityModel::Start (void)
{
  m_helper.Update ();
  m_helper.Pause ();
  m_event = Simulator::Schedule (m_pause, &BonnMotionMobilityModel::BeginWalk, this);
  NotifyCourseChange ();
}
*/
Vector
BonnMotionMobilityModel::DoGetPosition (void) const
{
  m_helper.Update ();
  return m_helper.GetCurrentPosition ();
}

void 
BonnMotionMobilityModel::DoSetPosition (const Vector &position)
{
  m_helper.SetPosition (position);
  Simulator::Remove (m_event);
  m_event = Simulator::ScheduleNow (&BonnMotionMobilityModel::BeginWalk, this);
}

Vector
BonnMotionMobilityModel::DoGetVelocity (void) const
{
  return m_helper.GetVelocity ();
}

} // namespace ns3
